"""File-system Transport.

Transport using the file-system as the message store.
"""
from __future__ import absolute_import, unicode_literals

import os
import shutil
import uuid
import re

from . import virtual
from kombu.exceptions import ChannelError
from kombu.five import Empty, monotonic
from kombu.utils.encoding import bytes_to_str, str_to_bytes
from kombu.utils.json import loads, dumps
from kombu.utils.objects import cached_property


VERSION = (1, 0, 0)
__version__ = '.'.join(map(str, VERSION))


class FileLock(object):
    def __init__(self, filename):
        self._lock_file = filename + '.lock'
        self._lock_file_fd = None

    def acquire(self):
        open_mode = os.O_WRONLY | os.O_CREAT | os.O_EXCL | os.O_TRUNC
        try:
            fd = os.open(self._lock_file, open_mode)
        except (IOError, OSError):
            return False
        else:
            self._lock_file_fd = fd
            return True

    def release(self):
        os.close(self._lock_file_fd)
        os.remove(self._lock_file)
        self._lock_file_fd = None


class Channel(virtual.Channel):
    """Filesystem Channel."""

    def _put(self, queue, payload, **kwargs):
        """Put `message` onto `queue`."""
        filename = '%s_%s.%s.msg' % (int(round(monotonic() * 1000)),
                                     uuid.uuid4(), queue)
        filename = os.path.join(self.data_folder_out, filename)
        lock = FileLock(filename)
        if lock.acquire():
            try:
                f = open(filename, 'wb')
                f.write(str_to_bytes(dumps(payload)))
            except (IOError, OSError):
                raise ChannelError(
                    'Cannot add file {0!r} to directory'.format(filename))
            finally:
                f.close()
                lock.release()

    def _get(self, queue):
        """Get next message from `queue`."""
        folder = [f for f in os.listdir(self.data_folder_in) if re.search('.' + queue + '.msg', f) and (not re.search('.lock', f))] 
        folder = sorted(folder)
        while len(folder) > 0:
            filename = folder.pop(0)
            filename = os.path.join(self.data_folder_in, filename)
            lock = FileLock(filename)
            if lock.acquire():
                if not os.path.isfile(filename):
                    lock.release()
                    continue             
                try:
                    f = open(filename, 'r+b')
                    payload = f.read()
                except (IOError, OSError):
                    f.close()
                    lock.release()
                    raise ChannelError(
                        'Cannot read file {0!r} from queue.'.format(filename))
                f.close()
                if self.store_processed:
                    shutil.copy2(filename,
                                 self.processed_folder)                
                os.remove(filename)
                lock.release()
                return loads(bytes_to_str(payload))             

        raise Empty()

    def _purge(self, queue):
        """Remove all messages from `queue`."""
        count = 0
        queue_find = '.' + queue + '.msg'

        folder = os.listdir(self.data_folder_in)
        while len(folder) > 0:
            filename = folder.pop()
            try:
                # only purge messages for the requested queue
                if filename.find(queue_find) < 0:
                    continue

                filename = os.path.join(self.data_folder_in, filename)
                os.remove(filename)

                count += 1

            except OSError:
                # we simply ignore its existence, as it was probably
                # processed by another worker
                pass

        return count

    def _size(self, queue):
        """Return the number of messages in `queue` as an :class:`int`."""
        count = 0

        queue_find = '.{0}.msg'.format(queue)
        folder = os.listdir(self.data_folder_in)
        while len(folder) > 0:
            filename = folder.pop()

            # only handle message for the requested queue
            if filename.find(queue_find) < 0:
                continue

            count += 1

        return count

    @property
    def transport_options(self):
        return self.connection.client.transport_options

    @cached_property
    def data_folder_in(self):
        return self.transport_options.get('data_folder_in', 'data_in')

    @cached_property
    def data_folder_out(self):
        return self.transport_options.get('data_folder_out', 'data_out')

    @cached_property
    def store_processed(self):
        return self.transport_options.get('store_processed', False)

    @cached_property
    def processed_folder(self):
        return self.transport_options.get('processed_folder', 'processed')


class Transport(virtual.Transport):
    """Filesystem Transport."""

    Channel = Channel

    default_port = 0
    driver_type = 'filesystem'
    driver_name = 'filesystem'

    def driver_version(self):
        return 'N/A'